const sync = require('./sync');
const murmurhash = require('murmurhash');
const cookie = require('cookie');
const url = require('url');
const uuidv4 = require('uuid/v4');
const logger = require('./logger');
const HASH_SEED = 1;
const MAX_HASH_VALUE = Math.pow(2, 32);
const MAX_TRAFFIC_VALUE = 10000;
const COOKIE_NAME = 'ABBY_EXPERIMENTS';
const CONTROL_VARIANT = 'control';
/**
* Helper function to generate bucket value in half-closed interval [0, MAX_TRAFFIC_VALUE)
* @param  {string} experimentID String value for experiment ID
* @return {number} the generated hash value
* @throws If bucketing value is not a valid string
*/
function generateHash(experimentID) {
    try {
        let hashValue = murmurhash.v3(experimentID, HASH_SEED);
        let ratio = hashValue / MAX_HASH_VALUE;
        return parseInt(ratio * MAX_TRAFFIC_VALUE, 10);
    } catch (ex) {
        logger.error(ex.message);
    }
}

/**
  * Helper function to generate bucket value in half-closed interval [0, MAX_TRAFFIC_VALUE)
  * @param  {Object} Experiment Object value for experiment
  * @param  {number} hash number value of generated hash
  * @return {string} the id of available variant
  */
function getVariation(experiment, hash, currentExperiment) {
    var percentage = Math.floor((hash * 100) / MAX_TRAFFIC_VALUE),
        variants = experiment.variants.filter(variant => {
            return percentage >= variant.range.start && percentage <= variant.range.end;
        });

    if (variants.length === 0) {
        return CONTROL_VARIANT;
    } else if (!currentExperiment || currentExperiment.variantCode === CONTROL_VARIANT) {
        return variants[0].code;
    } else {
        return currentExperiment.variantCode;
    }
}

/**
  * Helper function to generate bucket value in half-closed interval [0, MAX_TRAFFIC_VALUE)
  * @param  {Object} http.ServerResponse object value for http response
  * @param  {Array<String>} data array list of cookie data
  * @return {Void}
  */
function setCookie(response, data) {
    /**
     * set expires to 1 year
     */
    let path = '/',
        maxAge = 60 * 60 * 24 * 7 * 52,
        expires = new Date(Date.now() + maxAge);
    response.setHeader('Set-Cookie', cookie.serialize(COOKIE_NAME, data.join(','), {
        path,
        maxAge,
        expires
    }));
}

/**
  * Helper function to generate bucket value in half-closed interval [0, MAX_TRAFFIC_VALUE)
  * @param  {Object} http.ServerResponse object value for http response
  * @param  {Array<String>} data array list of cookie data
  * @return {Void}
  */
function getCookieData(req) {
    let data = req.headers.cookie ? cookie.parse(req.headers.cookie) : {};
    return data;
}

function getCurrentExperiments(data) {
    if (typeof data === 'string') {
        let values = data.split(',');
        let experiments = {};
        values.forEach(value => {
            let [experimentCode, variantCode, hash] = value.split('|');
            experiments[experimentCode] = { experimentCode, variantCode, hash };
        });
        return experiments;
    } else {
        return {};
    }
}

function getForcedExperimentsFromQuery(req) {
    const { query: { abby_experiment: query } } = url.parse(req.url, true);
    if (typeof query === 'string') {
        let values = query.split(',');
        let experiments = {};
        values.forEach(value => {
            let [experimentCode, variantCode] = value.split('|');
            experiments[experimentCode] = variantCode;
        });
        return experiments;
    } else {
        return {};
    }
}

module.exports = function (/**Configs*/configs, /**Synchroniser?*/synchroniser = sync) {
    let errors = [];
    if (!configs || typeof configs !== 'object') {
        errors.push(new Error('You should pass config object'));
    } else {
        if (!configs.tags) {
            errors.push(new Error('You should pass <tags>'));
        }
        if (!configs.apiEndpoint) {
            errors.push(new Error('You should pass <apiEndpoint>'));
        }
        if(configs.logger) {
            logger.setLogger(configs.logger);
        }
    }

    if (errors.length > 0) {
        errors.forEach(error => logger.error(error));
        return;
    }

    function handle(req, res) {
        let allExperiments = synchroniser.data,
            availableExperiments = {},
            cookieData = [],
            cookies = getCookieData(req),
            currentExperiments = getCurrentExperiments(cookies[COOKIE_NAME]),
            forcedExperiments = getForcedExperimentsFromQuery(req);

        for (let experiment of allExperiments) {
            let experimentCode = experiment.code,
                currentExperiment = currentExperiments[experimentCode],
                hash = currentExperiment && currentExperiment.hash ? currentExperiment.hash : generateHash(`${uuidv4()}_${experimentCode}`);

            availableExperiments[experimentCode] = forcedExperiments[experimentCode] || getVariation(experiment, hash, currentExperiment);
            cookieData.push(`${experimentCode}|${availableExperiments[experimentCode]}|${hash}`);
        }
        setCookie(res, cookieData);

        req.experiments = availableExperiments;
    }

    handle.ready = () => synchroniser.load(configs);
    return handle;
};
