const expect = require('chai').expect;
const assert = require('chai').assert;
const abby = require('../lib/client');
const mocks = require('../mocks/index');

describe('client', () => {
    let abbyClient,
        synchroniser = mocks.synchroniser;
    beforeEach(() => {
        abbyClient = abby({
            tags: 'test, production',
            apiEndpoint: 'https://abby.io/api/v1/'
        }, synchroniser);
    });

    describe('when there is no Abby cookie', () => {
        let request,
            response;
        beforeEach(() => {
            // create the request object having no cookies
            request = {
                headers: {
                    cookie: null
                },
                url: 'anything'
            };
            // create the response object for later reference
            response = {
                headers: {},
                setHeader() {
                    // do something
                    response.headers.cookie = 'uhhh cookies!';
                }
            };
            // return experiments from api
            abbyClient(request, response);
        });
        it('should set all features to request', () => {
            expect(request.experiments.feature_abc).to.not.undefined;
            expect(request.experiments.feature_def).to.not.undefined;
        });
        it('should store the evaluation result in Cookies', () => {
            // check res.cookies
            expect(response.headers.cookie).to.equal('uhhh cookies!');
        });
    });

    describe('when there are already an Abby cookie', () => {
        beforeEach(() => {
            // create the request object having cookies
            let request = {
                headers: {
                    cookie: '_ga=GA1.2.77836415.1509897097; _gid=GA1.2.54182306.1519757530; optimizelyBuckets=%7B%7D; optimizelySegments=%7B%7D; constructorio_t=djF8fDk0NDM0ODFiLWU2ODMtNDRiOC04YzMyLTAzM2RiYmEyMjM4NA==; crumb=E0_d9ZZjl6rPUYJu34W2tAHANRt2wOcKeNlQD5FpOhf; ConstructorioID_client_id=9443481b-e683-44b8-8c32-033dbba22384; optimizelyEndUserId=oeu1509897097017r0.49044477643947715'
                },
            };
            // create the response object for later reference
            let response = {
                headers: {},
                setHeader() {
                    // do something
                    response.headers.cookie = 'uhhh cookies!';
                }
            };
            // return experiments from api
            abbyClient(request, response);
        });
        // it('should put the client to the same non-control variation as set in the cookie', () => {
        //     //
        // });
    });

    describe('when there is a forced query experiment in request', () => {
        let request,
            response;
        function initSpec({isAbbyQuery = true, experiment = 'feature_ghi', variation = 'variation_1_ghi'} = {}) {
            // create the request object having no cookies
            request = {
                headers: {
                    cookie: null
                },
                url : isAbbyQuery ? `?abby_experiment=${experiment}|${variation}` : `anything`
            };
            // create the response object for later reference
            response = {
                headers: {},
                setHeader() {
                    // do something
                    response.headers.cookie = 'uhhh cookies!';
                }
            };
            // return experiments from api
            abbyClient(request, response);
        }
        it('should set feature_ghi to variation_1_ghi', () => {
            initSpec();
            expect(request.experiments.feature_ghi).to.equal('variation_1_ghi');
        });
    });

    describe('distribution of original/variation', () => {
        let deviation;

        function initSpec() {
            let originalCount = 0,
                variantCount = 0,
                distributionRate = 10000,
                originalPercent,
                variantPercent,
                request,
                response;
            // create the request object having no cookies
            request = {
                headers: {
                    cookie: null
                },
                url: 'anything'
            };
            // create the response object for later reference
            response = {
                headers: {},
                setHeader() {
                    // do something
                    response.headers.cookie = 'uhhh cookies!';
                }
            };
            for (let i = 0; i < distributionRate; i++) {
                abbyClient(request, response);
                request.experiments.feature_ghi === 'variation_1_ghi' ? variantCount++ : originalCount++;
            }
            originalPercent = (originalCount * 100) / distributionRate;
            variantPercent = (variantCount * 100) / distributionRate;
            deviation = Math.abs(originalPercent - variantPercent);
        }

        it('should have approximately same range', () => {
            initSpec();
            assert.isAtMost(deviation, 5);
        });
    });
});
